% Implementation of Perceptual Exposure Fusion
%
% The method is developed (and so is the code) over the standard exposure
% fusion described in [Mertens et al. ,2007]
% Additional details of Perceptual Exposure Fusion in:
% Corneliu Florea, Constantin Vertan, Laura Florea
% "HIGH DYNAMIC RANGE IMAGING BY PERCEPTUAL LOGARITHMIC EXPOSURE MERGING"
% available at http://imag.pub.ro/common/staff/cflorea/cf_research.html/LIP
% Image Processing and Anlaysis Laboratory
% University Politehnica of Bucharest
% 
% The standard exposure fusion is described in [Mertens et al. ,2007]:
% Tom Mertens, Jan Kautz and Frank Van Reeth "Exposure Fusion"
%    In Proceedings of Pacific Graphics 2007
% and it was written by Tom Mertens, Hasselt University, August 2007
% 
%
% Usage:
%   result = PerceptExposureFusion(I,m);
%   Arguments:
%     'I': represents a stack of N color images (at double
%       precision). Dimensions are (height x width x 3 x N).
%     'm': 3-tuple that controls the per-pixel measures. The elements 
%     control contrast, saturation and well-exposedness, respectively.
%
% Example:
%   'figure; imshow(PerceptExposureFusion(I, [1 1 1]);'
%   This displays the fusion of the images in 'I' using fully calibrated
%   algortihm
%

function R = PerceptExposureFusion(I,m, USE_LTIP, logParam)

[nR,nC, nCh] = size(I);

N = size(I,4);

W = ones(nR,nCh,N);

%compute the measures and combines them into a weight map
contrast_parm = m(1);
sat_parm = m(2);
wexp_parm = m(3);

if (contrast_parm > 0)
    W = W.* contrast(I).^contrast_parm;
end
if (sat_parm > 0)
    W = W.*saturation(I).^sat_parm;
end
if (wexp_parm > 0)
    W = W.*well_exposedness(I).^wexp_parm;
end

%normalize weights: make sure that weights sum to one for each pixel
W = W + 1e-12; %avoids division by zero
W = W./repmat(sum(W,3),[1 1 N]);


% create empty pyramid
pyr = gaussian_pyramid(zeros(r,c,3));
nlev = length(pyr);

% multiresolution blending
for i = 1:N
    % construct pyramid from each input image
	pyrW = gaussian_pyramid(W(:,:,i));
	pyrI = laplacian_pyramid(I(:,:,:,i));
    
    % blend
    for l = 1:nlev
        w = repmat(pyrW{l},[1 1 3]);
        % composal using LTIP model
        pyr{l} = log_sumVert(pyr{l}, log_lambdaVert(pyrI{l},w,1),  1 ); ;
    end
end

% reconstruct
R = reconstruct_laplacian_pyramid(pyr);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
    

function im_result=log_sumVert(im1,im2,M);
%Usage : im_result=log_sumVert(im1,im2,M);
% where im1, im2 are image matrices
% M - is the maximum image value (e.g. 1 for double images or 255 for
% uint8)
% 
% It implements addition, pixel-wise, in logarithmic-type form  ( Vertan model),
% Input images (im1 and im2) values are in [0,M] 
% Output values (im_result) are in [0,M]

im1= im1./M;
im2= im2./M;
im_result=M.*(1-(1-im1).*(1-im2)./(1-im1.*im2));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   


function im_result=log_lambdaVert(im,lambda,M);
%Usage : im_result=log_lambdaVert(im,lambda,M);
% where:
%       im1 is 1-channel image matrix, lambda  
%       lambda - are sclar. It may be a unique scalar, or it may have the
%       same size as the image im1 so to facilitate convex combinations
% 
% It implements scalar multiplication, pixel-wise, in logarithmic-type form  ( Vertan model),
% Input image (im1) values are in [0,M] 
% Output values (im_result) are in [0,M]
im=im./M;
im_result=M.*lambda.*im./(1+(lambda-1).*im);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%--------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% contrast measure
function C = contrast(I)
h = [0 1 0; 1 -4 1; 0 1 0]; % laplacian filter
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    mono = rgb2gray(I(:,:,:,i));
    C(:,:,i) = (imfilter(mono,h,'replicate')).^0.25;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%--------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% saturation measure
function C = saturation(I)
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    % saturation is computed as the standard deviation of the color channels
    R = I(:,:,1,i);
    G = I(:,:,2,i);
    B = I(:,:,3,i);
    mu = (R + G + B)/3;
    C(:,:,i) = (((R - mu).^2 + (G - mu).^2 + (B - mu).^2)/3).^0.5

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%--------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % well-exposedness measure
function C = well_exposedness(I)
sig = .0.08;
N = size(I,4);
C = zeros(size(I,1),size(I,2),N);
for i = 1:N
    R = exp(-.5*(I(:,:,1,i) - 0.36).^2/sig.^2);
    G = exp(-.5*(I(:,:,2,i) - .37).^2/sig.^2);
    B = exp(-.5*(I(:,:,3,i) - 0.37).^2/sig.^2);
    C(:,:,i) = R.*G.*B;
end


